/*
 *      lark-core.c
 *
 *      Copyright 2016-2018 Alex <alex@linuxonly.ru>
 *
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>

#include "lark-core.h"

enum {
	LARK_CORE_TTS_RHVOICE = 0,
	LARK_CORE_TTS_ESPEAK
};

#define LARK_QUEUE_HIGH_LEVEL_BYTES 10*1024*1024
#define LARK_QUEUE_LOW_LEVEL_BYTES             0

#define LARK_QUEUE_HIGH_LEVEL_SEC             60
#define LARK_QUEUE_LOW_LEVEL_SEC              20

typedef struct lark_queue_node {
	guint8 *data;
	gsize length;
	struct lark_queue_node *next;
} lark_queue_node_t;

struct _LarkCore
{
	GObject parent_instance;
	/*Objects*/
	LarkGstreamer *gstreamer;
	LarkRHVoice *rhvoice;
	LarkEspeak *espeak;
	GSettings *prefsettings;
	GSettings *eqsettings;
	GSettings *expsettings;
	/*Data*/
	GThread *prodthread;
	GThread *consthread;
	GMutex mutex;
	GCond produce;
	GCond consume;
	gboolean finished;
	lark_queue_node_t *queuehead;
	lark_queue_node_t *queuetail;
	guint queuemin;
	guint queuemax;
	gsize queuelength;
	guint queuebytes;
	guint numsamples;
	gint tts;
	/*Signal handlers*/
	gulong rsampleshandler;
	gulong esampleshandler;
	gulong gstmsghandler;
};

struct _LarkCoreClass
{
	GObjectClass parent_class;
};

G_DEFINE_TYPE (LarkCore, lark_core, G_TYPE_OBJECT)

enum {
	PROP_0,
	PROP_GSTREAMER,
	PROP_RHVOICE,
	PROP_ESPEAK,
	PROP_PREFERENCES_SETTINGS,
	PROP_EQUALIZER_SETTINGS,
	PROP_EXPORT_SETTINGS,
	PROP_QUEUE_MAX_BYTES,
	PROP_QUEUE_MIN_BYTES,
	LAST_PROP
};

static GParamSpec *properties[LAST_PROP];

enum {
	SIGNAL_TEXT_NEEDED,
	SIGNAL_FINISHED,
	LAST_SIGNAL
};

static guint signals[LAST_SIGNAL];


static void lark_core_enqueue(LarkCore *self, guint8 *data, gsize length)
{
	lark_queue_node_t *newnode;
	
	if ((self == NULL) || (data == NULL) || (length == 0)) return;
	
	newnode = (lark_queue_node_t *)g_malloc0(sizeof(struct lark_queue_node));
	
	newnode->data = data;
	newnode->length = length;
	newnode->next = NULL;
	
	if ((self->queuehead == NULL) && (self->queuetail == NULL)) {
		self->queuehead = newnode;
		self->queuetail = newnode;
		self->queuelength++;
		self->queuebytes += length;
		return;
	}
	
	self->queuetail->next = newnode;
	self->queuetail = newnode;
	self->queuelength++;
	self->queuebytes += length;
}

static guint8 *lark_core_dequeue(LarkCore *self, gsize *length)
{
	lark_queue_node_t *newnode;
	guint8 *data;
	gsize datalength;
	
	if (self->queuehead == NULL) return NULL;
	
	newnode = self->queuehead;
	
	data = self->queuehead->data;
	datalength = self->queuehead->length;
	
	if (self->queuehead == self->queuetail) {
		self->queuehead = NULL;
		self->queuetail = NULL;
	} else {
		self->queuehead = self->queuehead->next;
	}
	
	if (length != NULL) *length = datalength;
	
	self->queuelength--;
	self->queuebytes -= datalength;
	
	g_free(newnode);
	
	return data;
}

static gpointer lark_core_producer_thread(LarkCore *self)
{
	gchar *string;
	
	g_return_val_if_fail(self != NULL, NULL);
	
	g_debug("Core producer thread started");
	
	while (!self->finished) {
		g_mutex_lock(&self->mutex);
		
		if ((self->queuebytes >= self->queuemax) && (!self->finished)) {
			g_cond_wait(&self->produce, &self->mutex);
		}
		
		/*Produce*/
		g_signal_emit_by_name(G_OBJECT(self), "text-needed", &string);
		if (string != NULL) {
			if (self->tts == LARK_CORE_TTS_RHVOICE) {
				lark_rhvoice_generate_samples(self->rhvoice, string, &self->finished);
			} else if (self->tts == LARK_CORE_TTS_ESPEAK) {
				lark_espeak_generate_samples(self->espeak, string, &self->finished);
			}
		} else {
			self->finished = TRUE;
		}
						
		g_cond_signal(&self->consume);
		g_mutex_unlock(&self->mutex);
	}
	
	g_debug("Core producer thread terminated");
	
	return NULL;
}

static gpointer lark_core_consumer_thread(LarkCore *self)
{
	guint8 *data;
	gsize length;
	gsize rounddata;
	
	g_return_val_if_fail(self != NULL, NULL);
	
	g_debug("Core consumer thread started");
	
	while (TRUE) {
		g_mutex_lock(&self->mutex);
		
		if ((self->queuebytes <= self->queuemin) && (!self->finished)) {
			g_cond_wait(&self->consume, &self->mutex);
		}
		
		rounddata = 0; 
		
		/*Consume*/
		data = lark_core_dequeue(self, &length);
		if (data != NULL) {
			if (!lark_gstreamer_add_data(self->gstreamer, data, length)) {
				/*After pipleine stop*/
				g_cond_signal(&self->produce);
				g_mutex_unlock(&self->mutex);
				break;
			} else {
				rounddata += length;
			}
		} else if (self->finished) {
			if (rounddata == 0) {
				/*Stop GStreamer by EOS*/
				lark_gstreamer_stop(self->gstreamer, FALSE);
			}
			/*After EOS*/
			g_cond_signal(&self->produce);
			g_mutex_unlock(&self->mutex);
			break;
		}
		
		g_cond_signal(&self->produce);
		g_mutex_unlock(&self->mutex);
	}
	
	g_debug("Core consumer thread terminated");
	
	return NULL;
}

static void lark_core_samples_generated(LarkTtsEngine *engine, guint count, gshort *samples, LarkCore *core)
{
	guint8 *data;
	gsize length;
	
	g_return_if_fail(engine != NULL);
	g_return_if_fail(count != 0);
	g_return_if_fail(samples != NULL);
	g_return_if_fail(core != NULL);
	
	length = count * sizeof(short);
	
	data = g_malloc0(length);
	
	memcpy(data, samples, length);
	
	lark_core_enqueue(core, data, length);
	
	core->numsamples += count;
}

static void lark_core_gstreamer_finished(LarkGstreamer *gstreamer, gboolean successfully, GError *error, LarkCore *core)
{
	g_return_if_fail(gstreamer != NULL);
	g_return_if_fail(core != NULL);
	
	/*Stop threads*/
	lark_core_stop(core);
	
	/*Pass signal*/
	g_signal_emit_by_name(G_OBJECT(core), "finished", successfully, error);
}

static void lark_core_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
{
	LarkCore *self = (LarkCore *)object;
	
	switch (prop_id) {
		case PROP_GSTREAMER:
			g_value_set_pointer(value, self->gstreamer);
			break;
		case PROP_RHVOICE:
			g_value_set_pointer(value, self->rhvoice);
			break;
		case PROP_ESPEAK:
			g_value_set_pointer(value, self->espeak);
			break;
		case PROP_PREFERENCES_SETTINGS:
			g_value_set_pointer(value, self->prefsettings);
			break;
		case PROP_EQUALIZER_SETTINGS:
			g_value_set_pointer(value, self->eqsettings);
			break;
		case PROP_EXPORT_SETTINGS:
			g_value_set_pointer(value, self->expsettings);
			break;
		case PROP_QUEUE_MAX_BYTES:
			g_value_set_uint(value, self->queuemax);
			break;
		case PROP_QUEUE_MIN_BYTES:
			g_value_set_uint(value, self->queuemin);
			break;
	}
}

static void lark_core_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
{
	LarkCore *self = (LarkCore *)object;
	
	switch (prop_id) {
		case PROP_GSTREAMER:
			self->gstreamer = g_value_get_pointer(value);
			break;
		case PROP_RHVOICE:
			self->rhvoice = g_value_get_pointer(value);
			break;
		case PROP_ESPEAK:
			self->espeak = g_value_get_pointer(value);
			break;
		case PROP_PREFERENCES_SETTINGS:
			self->prefsettings = g_value_get_pointer(value);
			break;
		case PROP_EQUALIZER_SETTINGS:
			self->eqsettings = g_value_get_pointer(value);
			break;
		case PROP_EXPORT_SETTINGS:
			self->expsettings = g_value_get_pointer(value);
			break;
		case PROP_QUEUE_MAX_BYTES:
			self->queuemax = g_value_get_uint(value);
			break;
		case PROP_QUEUE_MIN_BYTES:
			self->queuemin = g_value_get_uint(value);
			break;
	}
}

static void lark_core_constructed(GObject *object)
{
	LarkCore *self = (LarkCore *)object;
	gchar *strsetting;
	
	/*Bind signals*/
	self->esampleshandler = g_signal_connect(self->espeak, "samples-generated", (GCallback)lark_core_samples_generated, self);
	
	self->rsampleshandler = g_signal_connect(self->rhvoice, "samples-generated", (GCallback)lark_core_samples_generated, self);
	
	self->gstmsghandler = g_signal_connect(self->gstreamer, "finished", (GCallback)lark_core_gstreamer_finished, self);
	
	/*Restore equalizer settings*/
	lark_gstreamer_set_eq_gain(self->gstreamer, LARK_GSTREAMER_EQ_FREQ_100, g_settings_get_double(self->eqsettings, "gain-100hz"));
	lark_gstreamer_set_eq_gain(self->gstreamer, LARK_GSTREAMER_EQ_FREQ_300, g_settings_get_double(self->eqsettings, "gain-300hz"));
	lark_gstreamer_set_eq_gain(self->gstreamer, LARK_GSTREAMER_EQ_FREQ_1K,  g_settings_get_double(self->eqsettings, "gain-1khz"));
	lark_gstreamer_set_eq_gain(self->gstreamer, LARK_GSTREAMER_EQ_FREQ_3K,  g_settings_get_double(self->eqsettings, "gain-3khz"));
	lark_gstreamer_set_eq_gain(self->gstreamer, LARK_GSTREAMER_EQ_FREQ_8K,  g_settings_get_double(self->eqsettings, "gain-8khz"));
	
	/*Restore export settings*/
	lark_gstreamer_set_target(self->gstreamer, LARK_GSTREAMER_TARGET_WAV);
	
	strsetting = g_settings_get_string(self->expsettings, "file-format");
	if (strsetting != NULL) {
		if (g_strcmp0(strsetting, "wav") == 0) {
			lark_gstreamer_set_target(self->gstreamer, LARK_GSTREAMER_TARGET_WAV);
		} else if ((g_strcmp0(strsetting, "ogg-vorbis") == 0) && (lark_gstreamer_get_vorbis_available(self->gstreamer))) {
			lark_gstreamer_set_target(self->gstreamer, LARK_GSTREAMER_TARGET_VORBIS);
		} else if ((g_strcmp0(strsetting, "ogg-opus") == 0) && (lark_gstreamer_get_opus_available(self->gstreamer))) {
			lark_gstreamer_set_target(self->gstreamer, LARK_GSTREAMER_TARGET_OPUS);
		} else if ((g_strcmp0(strsetting, "flac") == 0) && (lark_gstreamer_get_flac_available(self->gstreamer))) {
			lark_gstreamer_set_target(self->gstreamer, LARK_GSTREAMER_TARGET_FLAC);
		} else if ((g_strcmp0(strsetting, "mp3") == 0) && (lark_gstreamer_get_mp3_available(self->gstreamer))) {
			lark_gstreamer_set_target(self->gstreamer, LARK_GSTREAMER_TARGET_MP3);
		} else if ((g_strcmp0(strsetting, "aac") == 0) && (lark_gstreamer_get_aac_available(self->gstreamer))) {
			lark_gstreamer_set_target(self->gstreamer, LARK_GSTREAMER_TARGET_AAC);
		}
		g_free(strsetting);
	}
	
	lark_gstreamer_set_quality(self->gstreamer, g_settings_get_uint(self->expsettings, "encoding-quality"));
	
	strsetting = g_settings_get_string(self->expsettings, "export-directory");
	if (strsetting != NULL) {
		if (g_file_test(strsetting, G_FILE_TEST_IS_DIR)) {
			lark_gstreamer_set_path(self->gstreamer, strsetting);			
		} else {
			lark_gstreamer_set_path(self->gstreamer, (gchar *)g_get_user_special_dir(G_USER_DIRECTORY_MUSIC));
		}
		g_free(strsetting);
	}
	
	strsetting = g_settings_get_string(self->expsettings, "split-mode");
	if (strsetting != NULL) {
		if (g_strcmp0(strsetting, "size") == 0) {
			lark_gstreamer_set_limit_mode(self->gstreamer, LARK_GSTREAMER_LIMIT_MODE_SIZE);
		} else if (g_strcmp0(strsetting, "time") == 0) {
			lark_gstreamer_set_limit_mode(self->gstreamer, LARK_GSTREAMER_LIMIT_MODE_TIME);
		} else {
			lark_gstreamer_set_limit_mode(self->gstreamer, LARK_GSTREAMER_LIMIT_MODE_NONE);
		}
		g_free(strsetting);
	}
	
	lark_gstreamer_set_size_limit(self->gstreamer, g_settings_get_uint(self->expsettings, "size-limit"));
	
	lark_gstreamer_set_time_limit(self->gstreamer, g_settings_get_uint(self->expsettings, "time-limit"));
	
	G_OBJECT_CLASS(lark_core_parent_class)->constructed(object);
}

static void lark_core_finalize(GObject *object)
{
	LarkCore *self = (LarkCore *)object;
	
	/*Unbind signals*/
	if (self->esampleshandler != 0) {
		g_signal_handler_disconnect(self->espeak, self->esampleshandler);
	}
	
	if (self->rsampleshandler != 0) {
		g_signal_handler_disconnect(self->rhvoice, self->rsampleshandler);
	}
	
	if (self->gstmsghandler != 0) {
		g_signal_handler_disconnect(self->gstreamer, self->gstmsghandler);
	}
	
	G_OBJECT_CLASS(lark_core_parent_class)->finalize(object);
}

static void lark_core_class_init(LarkCoreClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS(klass);
	
	object_class->get_property = lark_core_get_property;
	object_class->set_property = lark_core_set_property;
	
	object_class->constructed = lark_core_constructed;
	object_class->finalize = lark_core_finalize;
	
	properties[PROP_GSTREAMER] =
		g_param_spec_pointer("gstreamer",
							"Pointer to GStreamer object",
							"Pointer to GStreamer object",
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
	
	properties[PROP_RHVOICE] =
		g_param_spec_pointer("rhvoice",
							"Pointer to RHVoice object",
							"Pointer to RHVoice object",
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
	
	properties[PROP_ESPEAK] =
		g_param_spec_pointer("espeak",
							"Pointer to Espeak object",
							"Pointer to Espeak object",
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
							
	properties[PROP_PREFERENCES_SETTINGS] =
		g_param_spec_pointer("preferences-settings",
							"Preferences settings object",
							"Pointer to preferences settings object",
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
	
	properties[PROP_EQUALIZER_SETTINGS] =
		g_param_spec_pointer("equalizer-settings",
							"Equalizer settings object",
							"Pointer to equalizer settings object",
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
							
	properties[PROP_EXPORT_SETTINGS] =
		g_param_spec_pointer("export-settings",
							"Export settings object",
							"Pointer to export settings object",
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
	
	properties[PROP_QUEUE_MAX_BYTES] = 
		g_param_spec_uint("max-bytes",
							"Maximum queue length",
							"Maximum queue length in bytes",
							LARK_QUEUE_LOW_LEVEL_BYTES,
							LARK_QUEUE_HIGH_LEVEL_BYTES,
							LARK_QUEUE_LOW_LEVEL_BYTES,
							G_PARAM_READABLE);
	
	properties[PROP_QUEUE_MIN_BYTES] = 
		g_param_spec_uint("min-bytes",
							"Minimum queue length",
							"Minimum queue length in bytes",
							LARK_QUEUE_LOW_LEVEL_BYTES,
							LARK_QUEUE_HIGH_LEVEL_BYTES,
							LARK_QUEUE_LOW_LEVEL_BYTES,
							G_PARAM_READABLE);
	
	g_object_class_install_properties(object_class, LAST_PROP, properties);
	
	signals[SIGNAL_TEXT_NEEDED] =
		g_signal_new("text-needed",
							G_TYPE_FROM_CLASS(klass),
							G_SIGNAL_RUN_LAST | G_SIGNAL_MUST_COLLECT,
							0,
							NULL,
							NULL,
							g_cclosure_marshal_generic,
							G_TYPE_STRING,
							0,
							G_TYPE_NONE);
	
	signals[SIGNAL_FINISHED] =
		g_signal_new("finished",
							G_TYPE_FROM_CLASS(klass),
							G_SIGNAL_RUN_FIRST,
							0,
							NULL,
							NULL,
							g_cclosure_marshal_generic,
							G_TYPE_NONE,
							2,
							G_TYPE_BOOLEAN,
							G_TYPE_ERROR);
	
}

guint lark_core_get_queue_max_bytes(LarkCore *self)
{
	g_return_val_if_fail(self != NULL, 0);
	
	return self->queuemax;	
}

guint lark_core_get_queue_min_bytes(LarkCore *self)
{
	g_return_val_if_fail(self != NULL, 0);
	
	return self->queuemin;
}

static void lark_core_calculate_queue_lengths(LarkCore *self, LarkTtsEngine *engine)
{
	guint bytespersample;
	
	g_return_if_fail(self != NULL);
	g_return_if_fail(engine != NULL);
	
	switch (lark_tts_engine_get_format(engine)) {
		case LARK_GSTREAMER_SRC_FORMAT_U8:
		case LARK_GSTREAMER_SRC_FORMAT_S8:
			bytespersample = 1;
			break;
		case LARK_GSTREAMER_SRC_FORMAT_U16LE:
		case LARK_GSTREAMER_SRC_FORMAT_U16BE:
		case LARK_GSTREAMER_SRC_FORMAT_S16LE:
		case LARK_GSTREAMER_SRC_FORMAT_S16BE:
			bytespersample = 2;
			break;
		case LARK_GSTREAMER_SRC_FORMAT_U32LE:
		case LARK_GSTREAMER_SRC_FORMAT_U32BE:
		case LARK_GSTREAMER_SRC_FORMAT_S32LE:
		case LARK_GSTREAMER_SRC_FORMAT_S32BE:
			bytespersample = 4;
			break;
		default:
			bytespersample = 2;
			break;
		}
		
		self->queuemin = lark_tts_engine_get_channels(engine) * lark_tts_engine_get_samplerate(engine) * bytespersample * LARK_QUEUE_LOW_LEVEL_SEC;
		self->queuemax = lark_tts_engine_get_channels(engine) * lark_tts_engine_get_samplerate(engine) * bytespersample * LARK_QUEUE_HIGH_LEVEL_SEC;
}

void lark_core_start(LarkCore *self, gboolean isexport)
{
	gchar *selprofile, **profile, *selstring;
	GSList *voices, *voicenode;
	gboolean profilefound;
	LarkTtsEngineVoice *voice;
	
	g_return_if_fail(self != NULL);
	
	selprofile = g_settings_get_string(self->prefsettings, "voice-profile");
	
	profile = g_strsplit(selprofile, "::", 2);
	
	if (profile != NULL) {
		if (g_strv_length(profile) == 2) {
			profilefound = FALSE;
			if (g_strcmp0(profile[0], "rhvoice") == 0) {
				/*RHVoice engine selected*/
				voices = lark_tts_engine_get_voices(LARK_TTS_ENGINE(self->rhvoice));
				/*Select voice*/
				for (voicenode = voices; voicenode->next != NULL; voicenode = voicenode->next) {
					voice = (LarkTtsEngineVoice *)voicenode->data;
					if (g_strcmp0(profile[1], voice->name) == 0) {
						lark_tts_engine_set_voice_profile(LARK_TTS_ENGINE(self->rhvoice), voice->name);
						profilefound = TRUE;
						break;
					}
				}
				/*If profile not found, select first one*/
				if (!profilefound) {
					voice = (LarkTtsEngineVoice *)g_slist_nth_data(voices, 0);
					lark_tts_engine_set_voice_profile(LARK_TTS_ENGINE(self->rhvoice), voice->name);
				}
				/*Set synthesis parameters*/
				lark_tts_engine_set_pitch(LARK_TTS_ENGINE(self->rhvoice), g_settings_get_double(self->prefsettings, "pitch"));
				lark_tts_engine_set_rate(LARK_TTS_ENGINE(self->rhvoice), g_settings_get_double(self->prefsettings, "rate"));
				lark_tts_engine_set_volume(LARK_TTS_ENGINE(self->rhvoice), g_settings_get_double(self->prefsettings, "volume"));
				
				selstring = g_settings_get_string(self->prefsettings, "capitals-symbols-indication");
				if (g_strcmp0(selstring, "pitch") == 0) {
					lark_tts_engine_set_capitals_indication(LARK_TTS_ENGINE(self->rhvoice), LARK_TTS_CAPITALS_SYMBOLS_INDICATION_PITCH);
				} else if (g_strcmp0(selstring, "volume") == 0) {
					lark_tts_engine_set_capitals_indication(LARK_TTS_ENGINE(self->rhvoice), LARK_TTS_CAPITALS_SYMBOLS_INDICATION_VOLUME);
				} else if (g_strcmp0(selstring, "off") == 0) {
					lark_tts_engine_set_capitals_indication(LARK_TTS_ENGINE(self->rhvoice), LARK_TTS_CAPITALS_SYMBOLS_INDICATION_OFF);
				} else {
					lark_tts_engine_set_capitals_indication(LARK_TTS_ENGINE(self->rhvoice), LARK_TTS_CAPITALS_SYMBOLS_INDICATION_DEFAULT);
				}
				g_free(selstring);
				
				selstring = g_settings_get_string(self->prefsettings, "punctuation-symbols-indication");
				if (g_strcmp0(selstring, "on") == 0) {
					lark_tts_engine_set_punctuation_indication(LARK_TTS_ENGINE(self->rhvoice), LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_ON);
				} else if (g_strcmp0(selstring, "off") == 0) {
					lark_tts_engine_set_punctuation_indication(LARK_TTS_ENGINE(self->rhvoice), LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_OFF);
				} else {
					lark_tts_engine_set_punctuation_indication(LARK_TTS_ENGINE(self->rhvoice), LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_DEFAULT);
				}
				g_free(selstring);
				/*Prepare GStreamer*/
				lark_gstreamer_set_format(self->gstreamer, lark_tts_engine_get_format(LARK_TTS_ENGINE(self->rhvoice)));
				lark_gstreamer_set_sample_rate(self->gstreamer, lark_tts_engine_get_samplerate(LARK_TTS_ENGINE(self->rhvoice)));
				lark_gstreamer_set_channels(self->gstreamer, lark_tts_engine_get_channels(LARK_TTS_ENGINE(self->rhvoice)));
				
				lark_core_calculate_queue_lengths(self, LARK_TTS_ENGINE(self->rhvoice));
				
				self->tts = LARK_CORE_TTS_RHVOICE;
			} else if (g_strcmp0(profile[0], "espeak") == 0) {
				/*ESpeak engine selected*/
				voices = lark_tts_engine_get_voices(LARK_TTS_ENGINE(self->espeak));
				/*Select voice*/
				for (voicenode = voices; voicenode->next != NULL; voicenode = voicenode->next) {
					voice = (LarkTtsEngineVoice *)voicenode->data;
					if (g_strcmp0(profile[1], voice->name) == 0) {
						lark_tts_engine_set_voice_profile(LARK_TTS_ENGINE(self->espeak), voice->name);
						profilefound = TRUE;
						break;
					}
				}
				/*If profile not found, select first one*/
				if (!profilefound) {
					voice = (LarkTtsEngineVoice *)g_slist_nth_data(voices, 0);
					lark_tts_engine_set_voice_profile(LARK_TTS_ENGINE(self->espeak), voice->name);
				}
				/*Set synthesis parameters*/
				lark_tts_engine_set_pitch(LARK_TTS_ENGINE(self->espeak), g_settings_get_double(self->prefsettings, "pitch"));
				lark_tts_engine_set_rate(LARK_TTS_ENGINE(self->espeak), g_settings_get_double(self->prefsettings, "rate"));
				lark_tts_engine_set_volume(LARK_TTS_ENGINE(self->espeak), g_settings_get_double(self->prefsettings, "volume"));
				
				selstring = g_settings_get_string(self->prefsettings, "capitals-symbols-indication");
				if (g_strcmp0(selstring, "pitch") == 0) {
					lark_tts_engine_set_capitals_indication(LARK_TTS_ENGINE(self->espeak), LARK_TTS_CAPITALS_SYMBOLS_INDICATION_PITCH);
				} else if (g_strcmp0(selstring, "volume") == 0) {
					lark_tts_engine_set_capitals_indication(LARK_TTS_ENGINE(self->espeak), LARK_TTS_CAPITALS_SYMBOLS_INDICATION_VOLUME);
				} else if (g_strcmp0(selstring, "off") == 0) {
					lark_tts_engine_set_capitals_indication(LARK_TTS_ENGINE(self->espeak), LARK_TTS_CAPITALS_SYMBOLS_INDICATION_OFF);
				} else {
					lark_tts_engine_set_capitals_indication(LARK_TTS_ENGINE(self->espeak), LARK_TTS_CAPITALS_SYMBOLS_INDICATION_DEFAULT);
				}
				g_free(selstring);
				
				selstring = g_settings_get_string(self->prefsettings, "punctuation-symbols-indication");
				if (g_strcmp0(selstring, "on") == 0) {
					lark_tts_engine_set_punctuation_indication(LARK_TTS_ENGINE(self->espeak), LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_ON);
				} else if (g_strcmp0(selstring, "off") == 0) {
					lark_tts_engine_set_punctuation_indication(LARK_TTS_ENGINE(self->espeak), LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_OFF);
				} else {
					lark_tts_engine_set_punctuation_indication(LARK_TTS_ENGINE(self->espeak), LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_DEFAULT);
				}
				g_free(selstring);
				/*Prepare GStreamer*/
				lark_gstreamer_set_format(self->gstreamer, lark_tts_engine_get_format(LARK_TTS_ENGINE(self->espeak)));
				lark_gstreamer_set_sample_rate(self->gstreamer, lark_tts_engine_get_samplerate(LARK_TTS_ENGINE(self->espeak)));
				lark_gstreamer_set_channels(self->gstreamer, lark_tts_engine_get_channels(LARK_TTS_ENGINE(self->espeak)));
				
				lark_core_calculate_queue_lengths(self, LARK_TTS_ENGINE(self->espeak));
				
				self->tts = LARK_CORE_TTS_ESPEAK;
			} else {
				/*Assume it's RHVoice engine*/
				voices = lark_tts_engine_get_voices(LARK_TTS_ENGINE(self->rhvoice));
				voice = (LarkTtsEngineVoice *)g_slist_nth_data(voices, 0);
				lark_tts_engine_set_voice_profile(LARK_TTS_ENGINE(self->rhvoice), voice->name);
				/*Set synthesis parameters*/
				lark_tts_engine_set_pitch(LARK_TTS_ENGINE(self->rhvoice), g_settings_get_double(self->prefsettings, "pitch"));
				lark_tts_engine_set_rate(LARK_TTS_ENGINE(self->rhvoice), g_settings_get_double(self->prefsettings, "rate"));
				lark_tts_engine_set_volume(LARK_TTS_ENGINE(self->rhvoice), g_settings_get_double(self->prefsettings, "volume"));
				
				selstring = g_settings_get_string(self->prefsettings, "capitals-symbols-indication");
				if (g_strcmp0(selstring, "pitch") == 0) {
					lark_tts_engine_set_capitals_indication(LARK_TTS_ENGINE(self->rhvoice), LARK_TTS_CAPITALS_SYMBOLS_INDICATION_PITCH);
				} else if (g_strcmp0(selstring, "volume") == 0) {
					lark_tts_engine_set_capitals_indication(LARK_TTS_ENGINE(self->rhvoice), LARK_TTS_CAPITALS_SYMBOLS_INDICATION_VOLUME);
				} else if (g_strcmp0(selstring, "off") == 0) {
					lark_tts_engine_set_capitals_indication(LARK_TTS_ENGINE(self->rhvoice), LARK_TTS_CAPITALS_SYMBOLS_INDICATION_OFF);
				} else {
					lark_tts_engine_set_capitals_indication(LARK_TTS_ENGINE(self->rhvoice), LARK_TTS_CAPITALS_SYMBOLS_INDICATION_DEFAULT);
				}
				g_free(selstring);
				
				selstring = g_settings_get_string(self->prefsettings, "punctuation-symbols-indication");
				if (g_strcmp0(selstring, "on") == 0) {
					lark_tts_engine_set_punctuation_indication(LARK_TTS_ENGINE(self->rhvoice), LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_ON);
				} else if (g_strcmp0(selstring, "off") == 0) {
					lark_tts_engine_set_punctuation_indication(LARK_TTS_ENGINE(self->rhvoice), LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_OFF);
				} else {
					lark_tts_engine_set_punctuation_indication(LARK_TTS_ENGINE(self->rhvoice), LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_DEFAULT);
				}
				g_free(selstring);
				/*Prepare GStreamer*/
				lark_gstreamer_set_format(self->gstreamer, lark_tts_engine_get_format(LARK_TTS_ENGINE(self->rhvoice)));
				lark_gstreamer_set_sample_rate(self->gstreamer, lark_tts_engine_get_samplerate(LARK_TTS_ENGINE(self->rhvoice)));
				lark_gstreamer_set_channels(self->gstreamer, lark_tts_engine_get_channels(LARK_TTS_ENGINE(self->rhvoice)));
				
				lark_core_calculate_queue_lengths(self, LARK_TTS_ENGINE(self->rhvoice));
				
				self->tts = LARK_CORE_TTS_RHVOICE;
			}
		}
		g_strfreev(profile);
	}
	
	if (!isexport) {
		lark_gstreamer_set_target(self->gstreamer, LARK_GSTREAMER_TARGET_AUDIO);
	}
	
	lark_gstreamer_set_queue_bytes(self->gstreamer, self->queuemax);
	
	/*Start GStreamer*/
	lark_gstreamer_start(self->gstreamer);
	
	/*Create threads*/
	self->prodthread = g_thread_new("Lark producer thread", (GThreadFunc)lark_core_producer_thread, self);
	self->consthread = g_thread_new("Lark consumer thread", (GThreadFunc)lark_core_consumer_thread, self);
}

void lark_core_stop(LarkCore *self)
{
	guint8 *data;
	gchar *selstring;
	
	g_return_if_fail(self != NULL);
	
	if ((self->consthread != NULL) && (self->prodthread != NULL)) {
		/*Stop GStreamer by request*/
		lark_gstreamer_stop(self->gstreamer, TRUE);
		/*Set finished state*/
		self->finished = TRUE;
		/*Wait for threads termination*/
		g_thread_join(self->consthread);
		g_thread_join(self->prodthread);
	
		/*Free samples*/
		while (1) {
			data = lark_core_dequeue(self, NULL);
			if (data != NULL) {
				g_free(data);
			} else {
				break;
			}
		}
		/*Restore default settings*/
		g_mutex_init(&self->mutex);
		g_cond_init(&self->produce);
		g_cond_init(&self->consume);
		self->finished = FALSE;
		self->queuehead = NULL;
		self->queuetail = NULL;
		self->queuemin = 0;
		self->queuemax = 0;
		self->queuelength = 0;
		self->queuebytes = 0;
		self->numsamples = 0;
		self->consthread = NULL;
		self->prodthread = NULL;
	}
	
	/*Restore target*/
	lark_gstreamer_set_target(self->gstreamer, LARK_GSTREAMER_TARGET_WAV);
	selstring = g_settings_get_string(self->expsettings, "file-format");
	if (selstring != NULL) {
		if (g_strcmp0(selstring, "wav") == 0) {
			lark_gstreamer_set_target(self->gstreamer, LARK_GSTREAMER_TARGET_WAV);
		} else if ((g_strcmp0(selstring, "ogg-vorbis") == 0) && (lark_gstreamer_get_vorbis_available(self->gstreamer))) {
			lark_gstreamer_set_target(self->gstreamer, LARK_GSTREAMER_TARGET_VORBIS);
		} else if ((g_strcmp0(selstring, "ogg-opus") == 0) && (lark_gstreamer_get_opus_available(self->gstreamer))) {
			lark_gstreamer_set_target(self->gstreamer, LARK_GSTREAMER_TARGET_OPUS);
		} else if ((g_strcmp0(selstring, "flac") == 0) && (lark_gstreamer_get_flac_available(self->gstreamer))) {
			lark_gstreamer_set_target(self->gstreamer, LARK_GSTREAMER_TARGET_FLAC);
		} else if ((g_strcmp0(selstring, "mp3") == 0) && (lark_gstreamer_get_mp3_available(self->gstreamer))) {
			lark_gstreamer_set_target(self->gstreamer, LARK_GSTREAMER_TARGET_MP3);
		} else if ((g_strcmp0(selstring, "aac") == 0) && (lark_gstreamer_get_aac_available(self->gstreamer))) {
			lark_gstreamer_set_target(self->gstreamer, LARK_GSTREAMER_TARGET_AAC);
		}
		g_free(selstring);
	}
}

static void lark_core_init(LarkCore *self)
{
	/*Sync primitives*/
	g_mutex_init(&self->mutex);
	g_cond_init(&self->produce);
	g_cond_init(&self->consume);
	/*State variables*/
	self->finished = FALSE;
	self->queuehead = NULL;
	self->queuetail = NULL;
	self->queuemin = 0;
	self->queuemax = 0;
	self->queuelength = 0;
	self->queuebytes = 0;
	self->numsamples = 0;
}

LarkCore *lark_core_new(LarkGstreamer *gstreamer, LarkRHVoice *rhvoice, LarkEspeak *espeak, GSettings *prefsettings, GSettings *eqsettings, GSettings *expsettings)
{
	return g_object_new(lark_core_get_type(), "gstreamer", gstreamer, "rhvoice", rhvoice, "espeak", espeak, "preferences-settings", prefsettings, "equalizer-settings", eqsettings, "export-settings", expsettings, NULL);
}
